@@ -1,4 +1,4 @@
-/* $Id: daemonlogger.c,v 1.16 2007/11/09 20:47:30 roesch Exp $ */
+/* $Id: daemonlogger.c,v 1.20 2008/08/28 14:47:33 roesch Exp $ */
 
 /*************** IMPORTANT DAEMONLOGGER LICENSE TERMS **************** 
 * 
@@ -81,6 +81,10 @@
 ** Author: Martin Roesch <roesch@sourcefire.com>
 */
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
@@ -103,12 +107,31 @@
 #include <grp.h>
 #include <pwd.h>
 #include <ctype.h>
+#include <sys/queue.h>
+#include <sys/param.h>
+#include <sys/mount.h>
+
+#ifdef LINUX
+#include <sys/statvfs.h>
+#include <sys/vfs.h>
+#endif
 
 #define SUCCESS     0
 #define ERROR       1
 #define STDBUF      1024
+#define KILOBYTE    1 << 10
+#define MEGABYTE    1 << 20
 #define GIGABYTE    1 << 30
-#define VERSION     "1.0.1"
+#define TERABYTE    1 << 40
+
+#ifndef VERSION
+#define VERSION     "1.2.0"
+#endif
+
+#define _FILE_OFFSET_BITS   64
+
+#define PRUNE_OLDEST_ABSOLUTE   0
+#define PRUNE_OLDEST_IN_RUN     1
 
 typedef enum {
     MINUTES=1,
@@ -123,6 +146,33 @@ static char *interval_names[] = {
     "days"
 };
 
+typedef enum {
+    KILOBYTES = 1,
+    MEGABYTES,
+    GIGABYTES,
+    TERABYTES
+    } size;
+    
+static char *size_names[] = {
+    "none",
+    "kilobytes",
+    "megabytes",
+    "gigabytes",
+    "terabytes"
+};
+
+typedef struct filelist Filelist;
+
+struct file_entry
+{
+    TAILQ_ENTRY(file_entry) next;
+    char *filename;
+};
+
+TAILQ_HEAD(filelist, file_entry);
+
+static Filelist file_list;
+
 static int count;
 static int daemon_mode;
 static int rollover;
@@ -140,6 +190,8 @@ static int drop_privs_flag;
 static int chroot_flag;
 static int rollover_interval;
 static int flush_flag;
+static int maxpct;
+static int prune_flag;
 
 static char *interface;
 static char *retrans_interface;
@@ -154,20 +206,36 @@ static char *group_name;
 static char *user_name;
 static char *chroot_dir;
 static char logdir[STDBUF];
+static char testpath[STDBUF];
 
-static u_int32_t rollsize;
+static size_t rollsize;
 static time_t lastroll;
 static time_t nextroll;
+static u_int64_t rollsize_in_blocks;
 
 static pcap_t *pd;
 static pcap_dumper_t *pdp;
 
 static eth_t *eth_retrans;
 
+static u_int64_t part_total_blocks;
+static u_int64_t part_min_free_blocks;
+
 void (*packet_handler)(char *user, struct pcap_pkthdr *pkthdr, u_char *pkt);
 static int sniff_loop();
 static int set_rollover_time();
 
+#ifdef LINUX
+#define d_statfs(p, s) statvfs(p, s)
+typedef struct statvfs d_statfs_t; 
+#elif MACOSX
+#define d_statfs(p, s) statfs64(p, s)
+typedef struct statfs64 d_statfs_t;
+#else
+#define d_statfs(p, s) statfs(p, s)
+typedef struct statfs d_statfs_t;
+#endif
+
 static void fatal(const char *format, ...)
 {
     char buf[STDBUF+1];
@@ -333,7 +401,7 @@ char *get_filename()
                     "%s/%s.%lu", 
                     logpath, 
                     logfilename, 
-                    currtime) < 0)
+                    (long unsigned int) currtime) < 0)
             return NULL;        
     }
     else
@@ -342,7 +410,7 @@ char *get_filename()
                     STDBUF, 
                     "%s.%lu",
                     logfilename,
-                    currtime) < 0)
+                    (long unsigned int) currtime) < 0)
             return NULL;        
     }
 
@@ -385,7 +453,47 @@ static void quitter(int signal)
     alarm(1);
 }
 
-static int prune_oldest_file()
+static int prune_oldest_file_this_run()
+{
+    struct stat sb;
+    struct file_entry *fe;
+    
+    while((fe = TAILQ_FIRST(&file_list)) != NULL)
+    {
+        if(fe->filename != NULL)
+        {
+            if(stat(fe->filename, &sb) != 0)
+            {
+                msg("[ERR] stat failed for \"%s\": %s\n", fe->filename, 
+                    strerror(errno));
+                TAILQ_REMOVE(&file_list, fe, next);
+                free(fe->filename);
+                free(fe);                    
+            }
+            else
+            {
+                if((sb.st_mode & S_IFMT) == S_IFREG)
+                {
+                    msg("[!] Ringbuffer: deleting %s", fe->filename);
+                    unlink(fe->filename);
+                    TAILQ_REMOVE(&file_list, fe, next);
+                    free(fe->filename);
+                    free(fe);
+                    break;
+                }                        
+            }
+        }
+        else
+        {
+            TAILQ_REMOVE(&file_list, fe, next);
+            free(fe);
+        }
+    }
+    
+    return 0;
+}
+
+static int prune_oldest_file_in_dir()
 {
     DIR *dirp;
     struct dirent *dp;
@@ -440,8 +548,9 @@ static int prune_oldest_file()
 
 static int open_log_file()
 {
+    struct file_entry *fe;
     char *filepath = get_filename();
-
+    
     if(maxfiles == 0 || (maxfiles > 0 && filecount > 0))
     {
         if(maxfiles > 0)
@@ -461,14 +570,57 @@ static int open_log_file()
         }
         else
         {
-            prune_oldest_file();
+            if(prune_flag == PRUNE_OLDEST_IN_RUN)
+                prune_oldest_file_this_run();
+            else
+                prune_oldest_file_in_dir();
+        }
+    }
+    
+    if(maxpct != 0)
+    {
+        d_statfs_t s;
+        if(d_statfs(testpath, &s) != 0)
+        {
+            perror("Unable to stat partition!\n");
+            fatal("EPIC FAIL!");
+        }
+        
+        if((s.f_bavail - rollsize_in_blocks) < part_min_free_blocks)
+        {
+            msg("Disk max utilization reached, rolling over and pruning");
+            if(prune_flag == PRUNE_OLDEST_IN_RUN)
+                prune_oldest_file_this_run();
+            else
+                prune_oldest_file_in_dir();
         }
     }
     
     if(filepath != NULL)
     {
+        if(ringbuffer == 1)
+        {
+            fe = calloc(sizeof(struct file_entry), sizeof(char));
+            if((fe->filename = strdup(filepath)) != NULL)
+            {
+                if(prune_flag == PRUNE_OLDEST_IN_RUN)
+                    TAILQ_INSERT_TAIL(&file_list, fe, next);
+#ifdef DEBUG
+                msg("File_list contents:\n");
+                TAILQ_FOREACH(fe, &file_list, next)
+                {
+                    msg("   %s\n", fe->filename);             
+                }
+#endif
+            }
+            else
+            {
+                fatal("Lurene sez ur fucked\n");
+            }
+        }
+
         msg("Logging packets to %s", filepath);
-        if((pdp = pcap_dump_open(pd, filepath)) == NULL)    
+        if((pdp = pcap_dump_open(pd, filepath)) == NULL)
         {
             fatal("Unable to open log file %s\n", pcap_geterr(pd));
         }
@@ -790,6 +942,7 @@ void packet_dump(char *user, struct pcap_pkthdr *pkthdr, u_char *pkt)
         {
             if(lastroll + rollover < now)
             {
+                msg("Rollover timer has expired!");
                 log_rollover();
                 lastroll = now;    
             }            
@@ -798,6 +951,7 @@ void packet_dump(char *user, struct pcap_pkthdr *pkthdr, u_char *pkt)
         {
             if(now > nextroll)
             {
+                msg("Rollover timer has expired!");
                 log_rollover();
                 set_rollover_time();
             }
@@ -814,8 +968,11 @@ void packet_dump(char *user, struct pcap_pkthdr *pkthdr, u_char *pkt)
     if(flush_flag)
         pcap_dump_flush(pdp);
         
-    if((u_int32_t)ftell((FILE *) pdp) > rollsize)
+    if(((size_t)ftello((FILE *) pdp)) > rollsize)
     {
+        msg("Size limit reached (%zd - %zd = %zd), rolling over!", 
+            (size_t)ftell((FILE *) pdp), rollsize,
+            (size_t) ftell((FILE *) pdp) - rollsize);
         log_rollover();
     }
     
@@ -942,6 +1099,8 @@ static void usage()
     printf("        -i <intf>       Grab packets from interface <intf>\n");
     printf("        -l <path>       Log to directory <path>\n");
     printf("        -m <count>      Generate <count> log files and quit\n");
+    printf("        -M <pct>        In ringbuffer mode log data to <pct> of\n"
+           "                        volume capacity\n");
     printf("        -n <name>       Set output filename prefix to <name>\n");
     printf("        -o <outf>       Disable logging, retransmit data from\n"
            "                        <intf> to <outf>\n");
@@ -950,8 +1109,7 @@ static void usage()
     printf("        -r              Activate ringbuffer mode\n");
     printf("        -R <pcap file>  Read packets from <pcap file>\n");
     printf("        -s <bytes>      Rollover the log file every <bytes>\n");
-    printf("        -S <snaplen>    Set the number of bytes per packet to "
-           "capture to <snaplen>\n");
+    printf("        -S <snaplen>    Capture <snaplen> bytes per packet\n");
     printf("        -t <time>       Rollover the log file on time intervals\n");
     printf("        -u <user name>  Set user ID to <user name>\n");
     printf("        -v              Show daemonlogger version\n");
@@ -962,30 +1120,28 @@ extern int  optind, opterr, optopt;
 
 int parse_cmd_line(int argc, char *argv[])
 {
-    int ch;
-    char rollmetric;
-    char *endptr;
+    int ch = 0;
+    char rollmetric = 0;
+    char rollenum = 0;
+    size_t  rollpoint = 0;
+    char *endptr = NULL;
     char *bpf_filename = NULL;
     int bpf_file = 0;
 
     while((ch = getopt(argc, argv, 
-            "c:df:Fg:hi:l:m:n:o:p:P:rR:s:S:t:T:u:v"))!=-1)
+            "c:df:Fg:hi:l:m:M:n:o:p:P:rR:s:S:t:T:u:vz"))!=-1)
     {
         switch(ch)
         {
             case 'c':
                 count = atoi(optarg);
-                printf("[-] Configured to log %d packets\n", count);
                 break;
             case 'd':
                 daemon_mode = 1;
-                printf("[-] Daemon mode set\n");
                 break;
             case 'f':
                 bpf_filename = strdup(optarg);
                 bpf_file = 1;
-                printf("[-] Reading BPF filter in from file %s\n", 
-                        bpf_filename);
                 break;
             case 'F':
                 flush_flag = 1;
@@ -993,7 +1149,6 @@ int parse_cmd_line(int argc, char *argv[])
             case 'g':
                 group_name = strdup(optarg);
                 drop_privs_flag = 1;
-                printf("[-] Setting group ID to %s\n", group_name);
                 break;
             case 'h':
                 usage();
@@ -1001,73 +1156,95 @@ int parse_cmd_line(int argc, char *argv[])
                 break;
             case 'i':
                 interface = strdup(optarg);
-                printf("[-] Interface set to %s\n", interface);
                 break;
             case 'l':
                 logpath = strdup(optarg);
-                printf("[-] Logpath set to %s\n", logpath);
                 break;
             case 'm':
                 maxfiles = atoi(optarg);
                 filecount = maxfiles;
-                printf("[-] Max files to write set to %d\n", maxfiles);
+                break;
+            case 'M':
+                maxpct = atoi(optarg);
+                if(maxpct > 100 || maxpct < 0)
+                    fatal("Bad max percent argument: %s\n", optarg);
                 break;
             case 'n':
                 free(logfilename);
                 logfilename = strdup(optarg);
-                printf("[-] Log filename set to \"%s\"", logfilename);
                 break;
             case 'o':
                 retrans_interface = strdup(optarg);
                 packet_handler = packet_retrans;
-                printf("[-] Tap output interface set to %s", 
-                        retrans_interface);
+
                 break;
             case 'p':
                 pidfile = strdup(optarg);
-                printf("[-] Pidfile configured to \"%s\"\n", pidfile);
                 break;
             case 'P':
                 pidpath = strdup(optarg);
-                printf("[-] Pidpath configured to \"%s\"\n", pidpath);
                 break;
             case 'r':
                 ringbuffer = 1;
-                printf("[-] Ringbuffer active\n");
                 break;
             case 'R':
                 readback_mode = 1;
                 readfile = strdup(optarg);
-                printf("[-] In readback mode\n");
                 break;
             case 's':
-                rollsize = strtoul(optarg, &endptr, 10);    
-                if(endptr == optarg)
+                if(isdigit((int)optarg[strlen(optarg)-1]))
                 {
-                    fprintf(stderr, "Bad rollover size, defaulting to 2GB\n");
-                    rollsize = GIGABYTE;
+                    rollsize = strtoul(optarg, &endptr, 10);    
+                    if(endptr == optarg)
+                    {
+                        fprintf(stderr, "Bad rollover size, defaulting to 2GB\n");
+                        rollsize = GIGABYTE;
+                    }
+                }
+                else
+                {
+                    sscanf(optarg, "%zu%c", &rollpoint, &rollmetric);
+                    
+                    switch(tolower(rollmetric))
+                    {
+                        case 'k':
+                            rollenum = KILOBYTES;
+                            rollsize = rollpoint * KILOBYTE;
+                            break;
+                        case 'm':
+                            rollsize = rollpoint * MEGABYTE;
+                            rollenum = MEGABYTES;
+                            break;
+                        case 'g':
+                            rollsize = rollpoint * GIGABYTE;
+                            rollenum = GIGABYTES;
+                            break;
+                        case 't':
+                            rollsize = (u_int64_t) rollpoint * TERABYTE;
+                            rollenum = TERABYTES;                            
+                            break;
+                        default:
+                            fatal("Bad size argument \"%c\"\n", 
+                                  rollmetric);
+                            break;
+                    }
                 }
-                printf("[-] Rollover size set to %lu bytes\n", 
-                        (unsigned long) rollsize);
                 break;
             case 'S':
                 if(!isdigit(optarg[0]))
                     fatal("Bad snaplen argument \"%s\"\n", optarg);
                 snaplen = atoi(optarg);
-                printf("[-] Snaplen set to %d\n", snaplen);
                 break;
             case 't':
                 if(isdigit((int)optarg[strlen(optarg)-1]))
                 {
                      rollover = atoi(optarg); 
-                     printf("[-] Rollover time configured for %d seconds\n", 
-                             rollover);
                 }
                 else
                 {
                     sscanf(optarg, "%d%c", &rollover, &rollmetric);
                     
-                    switch(rollmetric)
+                    switch(tolower(rollmetric))
                     {
                         case 'm':
                             rollover_interval = MINUTES;
@@ -1083,23 +1260,22 @@ int parse_cmd_line(int argc, char *argv[])
                                   rollmetric);
                             break;
                     }
-                    printf("[-] Rollover configured for %d %s\n", 
-                            rollover, interval_names[rollover_interval]);
                 }
                 break;
             case 'T':
                 chroot_dir = strdup(optarg);
                 chroot_flag = 1;
-                printf("[-] Setting chroot directory to %s", chroot_dir);
                 break;
             case 'u':
                 user_name = strdup(optarg);
                 drop_privs_flag = 1;
-                printf("[-] Setting user ID to %s\n", user_name);
                 break;
             case 'v':
                 showver = 1;
                 break;
+            case 'z':
+                prune_flag = PRUNE_OLDEST_IN_RUN;
+                break;
             default:
                 break;          
         }
@@ -1109,7 +1285,61 @@ int parse_cmd_line(int argc, char *argv[])
         pcap_cmd = copy_argv(&argv[optind]);
     else
         pcap_cmd = load_bpf_file(bpf_filename);
-        
+    
+    if(ringbuffer == 1 && prune_flag == PRUNE_OLDEST_IN_RUN)
+        TAILQ_INIT(&file_list);
+    
+    if(count)
+        printf("[-] Configured to log %d packets\n", count);
+    if(daemon_mode)
+        printf("[-] Daemon mode set\n");
+    if(bpf_file)
+        printf("[-] Reading BPF filter in from file %s\n", bpf_filename);
+    if(flush_flag)
+        printf("[-] Packet-buffered output activated\n");
+    if(drop_privs_flag)
+        printf("[-] Setting group ID to %s\n", group_name);
+    if(interface != NULL)
+        printf("[-] Interface set to %s\n", interface);
+    if(logpath != NULL)
+        printf("[-] Logpath set to %s\n", logpath);
+    if(filecount)
+        printf("[-] Max files to write set to %d\n", maxfiles);
+    if(logfilename != NULL)
+        printf("[-] Log filename set to \"%s\"\n", logfilename);
+    if(retrans_interface != NULL)
+        printf("[-] Tap output interface set to %s", retrans_interface);
+    if(pidfile)
+        printf("[-] Pidfile configured to \"%s\"\n", pidfile);
+    if(pidpath)
+        printf("[-] Pidpath configured to \"%s\"\n", pidpath);
+    if(ringbuffer)
+        printf("[-] Ringbuffer active\n");
+    if(readback_mode)
+        printf("[-] In readback mode\n");
+    if(rollpoint != 0)
+        printf("[-] Rollover configured for %zu %s\n", 
+                rollpoint, size_names[(int)rollenum]);
+    else
+        printf("[-] Rollover size set to %lu bytes\n", 
+                (unsigned long) rollsize);
+    if(snaplen)
+        printf("[-] Snaplen set to %d\n", snaplen);
+    if(rollmetric != 0)
+        printf("[-] Rollover configured for %d %s\n", 
+                rollover, interval_names[rollover_interval]);
+    else
+        printf("[-] Rollover time configured for %d seconds\n", 
+                rollover);
+    if(chroot_flag)
+        printf("[-] Setting chroot directory to %s", chroot_dir);
+    if(drop_privs_flag)
+        printf("[-] Setting user ID to %s\n", user_name);
+    if(prune_flag == PRUNE_OLDEST_IN_RUN)
+        printf("[-] Pruning behavior set to oldest THIS RUN\n");
+    else
+        printf("[-] Pruning behavior set to oldest IN DIRECTORY\n");
+
     return SUCCESS;
 }
 
@@ -1117,6 +1347,8 @@ int main(int argc, char *argv[])
 {   
     sigset_t set;
     packet_handler = packet_dump;
+    int statret;
+    d_statfs_t s;
     
     sigemptyset(&set);
     sigprocmask(SIG_SETMASK, &set, NULL);
@@ -1132,23 +1364,51 @@ int main(int argc, char *argv[])
     signal(SIGHUP, restarter);   if(errno!=0) errno=0;
     signal(SIGALRM, dl_shutdown);  if(errno!=0) errno=0;
     
-    rollsize = GIGABYTE;
+    rollsize = 2*GIGABYTE;
 
     logfilename = strdup("daemonlogger.pcap");
 
     parse_cmd_line(argc, argv);
 
     printf("\n-*> DaemonLogger <*-\n"
-           "Version %s\n\n"
+           "Version %s\n"
            "By Martin Roesch\n"
-           "(C) Copyright 2006-2007 Sourcefire Inc., All rights reserved\n"
+           "(C) Copyright 2006-2007 Sourcefire Inc., All rights reserved\n\n"
            , VERSION);    
 
     if(showver) exit(0);
 
     if(logpath != NULL && !is_valid_path(logpath))
         fatal("Log path \"%s\" is bad", logpath);
-
+        
+    if(logpath != NULL)
+    {
+        snprintf(testpath, STDBUF-1, "%s/.", logpath);
+        msg("Checking partition stats for log directory \"%s\"", testpath);
+        if((statret = d_statfs(testpath, &s)) != 0)
+        {
+            fatal("Unable to stat partition!\n\"%s\"\n", strerror(statret));
+        }
+        else
+        {
+            if(maxpct)
+            {
+                double pct;
+                double value;
+                
+                part_total_blocks = s.f_blocks;
+                pct = ((double) maxpct)/100.0;
+                value = ((double)part_total_blocks) * pct;
+                part_min_free_blocks = part_total_blocks - ((u_int64_t) value);
+                msg("%d%% max disk utilization = %llu blocks free (out of %llu)", 
+                    maxpct, part_min_free_blocks, part_total_blocks);
+                rollsize_in_blocks = (u_int64_t) (rollsize/(size_t)s.f_bsize);
+                msg("Blocksize = %lu", s.f_bsize);
+                msg("Rollsize = %llu blocks\n", rollsize_in_blocks);
+            }
+        }
+    }
+    
     if(daemon_mode) 
     {
         if(!is_valid_path(pidpath))
